
#include "format.h"

#if !defined(DEG2RAD)
  #define DEG2RAD 0.017453292519943
  #define RAD2DEG 57.29577951308232
#endif

#if !defined(MT2FT)
  #define MT2FT 3.28083333270619777
  #define FT2MT 0.30480060965948223
#endif

CStr::CStr() {
	len = 0;
	Allocate(len);
	ptr[0] = '\0';
}

CStr::CStr(const char *str) {
	len = strlen(str);
	Allocate(len);
	memcpy(ptr,str,len+1);
}

CStr::CStr(const char ch) {
	len = 1;
	Allocate(len);
	ptr[0] = ch;
	ptr[1] = '\0';
}

CStr::CStr(const CStr &str) {
	len = str.len;
	Allocate(len);
	memcpy(ptr,str.ptr,str.len+1);
}

CStr::CStr(const int l) {
	Allocate(l);
}

CStr::~CStr() {
	if (ptr) delete [] ptr;
	ptr = NULL;
}

CStr& CStr::operator=(const char ch) {
	delete [] ptr;
	len = 1;
	Allocate(len);
	ptr[0] = ch;
	ptr[1] = '\0';
	return *this;
}

CStr& CStr::operator=(const char *str) {
	delete [] ptr;
	len = strlen(str);
	Allocate(len);
	memcpy(ptr,str,len+1);
	return *this;
}

CStr& CStr::operator=(const CStr &str) {
	delete [] ptr;
	len = str.len;
	Allocate(len);
	memcpy(ptr,str.ptr,str.len+1);
	return *this;
}

CStr& CStr::operator+=(const char ch) {
	Reallocate(len+1);
	ptr[len++] = ch;
	ptr[len] = '\0';
	return *this;
}

CStr& CStr::operator+=(const char *str) {
	int length = strlen(str);
	int total = len + length;
	Reallocate(total);
	memcpy(&ptr[len],str,length+1);
	len = total;
	return *this;
}

CStr& CStr::operator+=(const CStr &str) {
	int total = len + str.len;
	Reallocate(total);
	memcpy(&ptr[len],str.ptr,str.len+1);
	len = total;
	return *this;
}

CStr CStr::operator+(const char *str) const {
	int length = strlen(str);
	int total = len + length;
	CStr tmp(total);
	memcpy(tmp.ptr,ptr,len);
	memcpy(&tmp.ptr[len],str,length+1);
	tmp.len = total;
	return tmp;
}

CStr CStr::operator+(const char ch) const {
	CStr tmp(len+1);
	tmp.len = len+1;
	memcpy(tmp.ptr,ptr,len);
	tmp.ptr[len] = ch;
	tmp.ptr[tmp.len] = '\0';
	return tmp;
}

CStr CStr::operator+(const CStr &str) const {
	int total = len + str.len;
	CStr tmp(total);
	memcpy(tmp.ptr,ptr,len);
	memcpy(&tmp.ptr[len],str.ptr,str.len+1);
	tmp.len = total;
	return tmp;
}

CStr operator+(const char *str1,const CStr &str2) {
	int length = strlen(str1);
	int total = length + str2.len;
	CStr tmp(total);
	memcpy(tmp.ptr,str1,length);
	memcpy(&tmp.ptr[length],str2.ptr,str2.len+1);
	tmp.len = total;
	return tmp;
}

bool CStr::operator==(const CStr &str) const {
	return (0==strcmp(ptr, str.ptr));
}

bool CStr::operator==(const char *str) const {
	return (0==strcmp(ptr, str));
}

bool CStr::operator!=(const CStr &str) const {
	return (0!=strcmp(ptr, str.ptr));
}

bool CStr::operator!=(const char *str) const {
	return (0!=strcmp(ptr, str));
}

CStr::operator const char* () const {
	return ptr;
}

const char* CStr::Ptr() const { return ptr; }

const int CStr::Length() const { return len; }

char* CStr::DynamicCopy() const {
	char *copy = new char[len+1];
	memcpy(copy,ptr,len+1);
	return copy;
};

void CStr::PadL(int min, char ch) {
	int chars = min-len;
	if (chars>0) {
		Reallocate(min);
		memmove(ptr+chars,ptr,len+1);
		memset(ptr,ch,chars);
		len = min;
	}
}

void CStr::Set(int start, int stop, char ch) {
	if (start>=stop || start>len-1 || stop>len || start<0 || stop<0) return;
	memset(ptr+start,ch,stop-start);
}

void CStr::Remove(char ch) {
	char *src = ptr;
	char *dst = ptr;
	while (*src) {
		if (*src==ch) {	src++; len--; }
		else *dst++ = *src++;
	}
	*dst = '\0';
}

void CStr::Allocate(int atleast) {
	max = (((atleast+1)+granularity-1)/granularity)*granularity-1;
	ptr = new char[max+1];
}

void CStr::Reallocate(int atleast) {
	if (atleast>max) {
		char *old = ptr;
		Allocate(atleast);
		memcpy(ptr,old,len+1);
		delete [] old;
	}
};

CStr CFormat::Double(double number,int frac,int min,char pad) {
	CStr output;
	if (frac < 0 || frac > 15) return output;

	double zeros = 1.0;
	double intpart; int i;
	for (i=0; i<frac; i++) zeros *= 10;

	if (number < 0.0) {
		number = -number;
		output += '-';
	}

	number += 0.5/zeros; // round
  
//	number = modf(number,&intpart);
	intpart = (long)number;	number -= intpart;

	output += Long((long)intpart,min,pad);
		
	if (frac) output += '.';

	for (i=0;i<frac;i++) {
//		number = modf(number*10,&intpart);
		number *= 10; intpart = (long)number; number -= intpart;
		output += (char)((long)intpart + '0');
	}
	return output;
}

CStr CFormat::Long(long number,int min,char pad) {
	CStr output;
	if (min < 0 || min > 15) return output;
	if (number < 0) output += '-';
  
	bool start = false;
	long i = 1000000000L;
	long remain = 10;
	while (remain--) {
		long r = number/i;
		number -= r*i;
		if (r<0) r = -r;
		start |= (r!=0) || (i==1);
		if (start) output += (char)r + '0';
		else if (remain<min) output += pad;
		i /= 10;
	}

	return output;
}

CStr CFormat::SLong(long number,int min,char pad) {
	CFormat format;
	return format.Long(number,min,pad);
}

CStr CFormat::SDouble(double number,int frac,int min,char pad) {
	CFormat format;
	return format.Double(number,frac,min,pad);
}

CFormatGPS::CFormatGPS() { Setup(); }

void CFormatGPS::Setup(DegFormat deg, SpeedFormat speed, DistFormat dist, DistFormat alt) {
	f_deg = deg;	f_speed = speed;
	f_dist = dist;	f_alt = alt;
}

CStr CFormatGPS::Degrees(double degrees,char positive,char negative) {
//	double dummy;
	char sign;
	if (degrees>=0) { sign = positive; }
	else { sign = negative; degrees = -degrees; }
//	double min = 60*modf(degrees,&dummy);
//	double sec = 60*modf(min,&dummy);
	double min = 60*(degrees-(int)degrees);
	double sec = 60*(min-(int)min);
	CStr tmp;
	switch (f_deg) {
		case degDDD:
			tmp += Double(degrees,6,3);
			tmp += '';
			break;
		case degDMM:
			tmp += Long((long)degrees,3);
			tmp += '';
			tmp += Double(min,3,2,'0');
			tmp += '\'';
			break;
		default:
			tmp += Long((long)degrees,3);
			tmp += '';
			tmp += Long((long)min,2,'0');
			tmp += '\'';
			tmp += Double(sec,1,2,'0');
			tmp += '"';
			break;
	}
	tmp += ' ';
	tmp += sign;
	return tmp;
}

CStr CFormatGPS::Latitude(double lat_d) {
	return Degrees(lat_d,'N','S');
}

CStr CFormatGPS::Longitude(double long_d) {
	return Degrees(long_d,'E','W');
}

CStr CFormatGPS::Northing(double northing) {
	return Double(northing,1);
}

CStr CFormatGPS::Easting(double easting) {
	return Double(easting,1);
}

CStr CFormatGPS::Speed(double speed) {
	switch (f_speed) {
		case spdMPS:
			return Double(speed,1,3);
		default:
			return Double(3.6*speed,1,3);
	}
}

CStr CFormatGPS::Dist(double dist) {
	switch (f_dist) {
		case dstFT:
			return Double(MT2FT*dist,1);
		default:
			return Double(dist,1);
	}
}

CStr CFormatGPS::Altitude(double dist) {
	switch (f_alt) {
		case dstFT:
			return Double(MT2FT*dist,1,4);
		default:
			return Double(dist,1,4);
	}
}

CStr CFormatGPS::Course(double course) {
	return Double(RAD2DEG*course,1,3,'0') + '';
}

CStr CFormatGPS::EPE(double epe) {
	CStr tmp('');
	tmp += Dist(epe);
	tmp.PadL(6,' ');
	return tmp;
}

CStr CFormatGPS::Time(time_t time) {
	const tm *t = gmtime(&time);
	return Long(t->tm_hour)+':'+Long(t->tm_min,2,'0')+':'+Long(t->tm_sec,2,'0');
}

CStr CFormatGPS::Date(time_t time) {
	const tm *t = gmtime(&time);
	return Long(t->tm_mday)+'.'+Long(t->tm_mon+1)+'.'+Long(t->tm_year);
}

CParse::CParse(const CStr &str, char delimiter) {
	Init(str,delimiter);
}

CParse::CParse(const char* str, char delimiter) {
	Init(str,delimiter);
}

void CParse::Init(const CStr &str, char delimiter) {
	ptr = str.Ptr();
	length = str.Length();
	default_delimiter = delimiter;
	ok = true;
}

void CParse::Init(const char* str, char delimiter) {
	ptr = str;
	length = strlen(str);
	default_delimiter = delimiter;
	ok = true;
}

bool CParse::Advance(int count) {
	if (count>length) {
		ptr += length;
		length = 0;
		return False();
	}
	else {
		ptr += count;
		length -= count;
		return true;
	}
}

void CParse::Spaces() {
	if (length == 0) return;
	int pos = 0;
	while (pos<length) {
		if (ptr[pos] != ' ') break;
		pos++;
	}	
	Advance(pos);
}

bool CParse::Delimiter(char delimiter) {
	Spaces();
	if ((length == 0) || (ptr[0] != delimiter)) return False();
	Advance(1);
	return true;
}

bool CParse::Delimiter() {
	return Delimiter(default_delimiter);
}

bool CParse::Character(char &ch) {
	if (length == 0) return False();
	ch = ptr[0];
	Advance(1);
	return true;
}

bool CParse::Const(const char *str) {
	int pos = 0;
	char act;
	while((act = str[pos])) {
		if (act!=ptr[pos]) return False();
		pos++;
	}
	Advance(pos);
	return true;
}

bool CParse::String(CStr &string, char delimiter) {
	string = "";
	int pos = 0;
	while (pos<length) {
		char act = ptr[pos];
		if (act == delimiter) break;
		string += act;
		pos++;
	}
	Advance(pos);
	return true;
}

bool CParse::String(CStr &string) {
	return String(string, default_delimiter);
}

bool CParse::Skip(char delimiter) {
	int pos = 0;
	while (pos<length) {
		if (ptr[pos] == delimiter) break;
		pos++;
	}
	Advance(pos);
	return true;
}

bool CParse::Skip() {
	return Skip(default_delimiter);
}

bool CParse::Long(long &number, int max) {
	Spaces();
	if (length==0) return False();
	if (max==0) max = length;

	long result = number = 0;
	int pos = 0;
	long sign = 1;
	if (ptr[0] == '-') {
		sign = -1;
		pos = 1;
	}

	bool valid = false;
	while (pos<max) {
		char act = ptr[pos];
		if (('0'<=act) && (act<='9')) {
			result = result * 10 + (act - '0');
			valid = true;
			pos++;
		}
		else break;
	}
	if (!valid) return False();
	number = result * sign;
	Advance(pos);
	return true;
}

bool CParse::Long(int &number, int max) {
	long number_long;
	bool ret = Long(number_long ,max);
	number = number_long;
	return ret;
}

bool CParse::Hex(long &number, int max) {
	Spaces();
	if (length==0) return False();
	if (max==0) max = length;

	long result = number = 0;
	int pos = 0;
	long sign = 1;
	if (ptr[0] == '-') {
		sign = -1;
		pos = 1;
	}

	bool valid = false;
	while (pos<max) {
		char act = ptr[pos];
		if (('0'<=act) && (act<='9')) {
			result = result * 16 + (act - '0');
			valid = true;
			pos++;
		}
		else if (('A'<=act) && (act<='F')) {
			result = result * 16 + (act - 'A' + 10);
			valid = true;
			pos++;
		}
		else break;
	}
	if (!valid) return False();
	number = result * sign;
	Advance(pos);
	return true;
}

bool CParse::Hex(int &number, int max) {
	long number_long;
	bool ret = Hex(number_long ,max);
	number = number_long;
	return ret;
}

bool CParse::Double(double &number) {
	Spaces();
	if (length==0) return False();

	double result = number = 0.0;
	int pos = 0;
	double sign = 1.0;
	if (ptr[0] == '-') {
		sign = -1.0;
		pos = 1;
	}

	bool valid = false;
	bool point = false;
	while (pos<length) {
		char act = ptr[pos];
		if (('0'<=act) && (act<='9')) {
			result = result * 10.0 + (act - '0');
			valid = true;
			pos++;
		}
		else if (act == '.') {
			point = true;
			pos++;
			break;
		}
		else break;
	}
	if (!valid) return False();
	if (point) {
		double fraction = 1;
		while (pos<length) {
			char act = ptr[pos];
			if (('0'<=act) && (act<='9')) {
				fraction /= 10.0;
				result += fraction * (act - '0');
				pos++;
			}
			else break;
		}
	}
	number = result * sign;
	Advance(pos);
	return true;
}

bool CParse::Double(float &number) {
	double number_double;
	bool ret = Double(number_double);
	number = (float)number_double;
	return ret;
}

bool CParse::Long(const char* str, long &number) {
	CParse parse(str);
	return (parse.Long(number) && parse.Empty());
}

bool CParse::Long(const char* str, int &number) {
	CParse parse(str);
	return (parse.Long(number) && parse.Empty());
}

bool CParse::Hex(const char* str, long &number) {
	CParse parse(str);
	return (parse.Hex(number) && parse.Empty());
}

bool CParse::Hex(const char* str, int &number) {
	CParse parse(str);
	return (parse.Hex(number) && parse.Empty());
}

bool CParseGPS::Latitude(double &lat_d) {
	return Degrees(lat_d, 'N', 'S');
}

bool CParseGPS::Longitude(double &long_d) {
	return Degrees(long_d, 'E', 'W');
}

bool CParseGPS::Latitude(const char* str, double &lat_d) {
	CParseGPS parse(str);
	return (parse.Latitude(lat_d) && parse.Empty());
}

bool CParseGPS::Longitude(const char* str, double &long_d) {
	CParseGPS parse(str);
	return (parse.Longitude(long_d) && parse.Empty());
}

bool CParseGPS::Degrees(double &degrees,char positive,char negative) {
	double deg, min, sec; deg = min = sec = 0;
	double sign = 1;

	if (Delimiter(positive)) sign = 1;
	if (Delimiter(positive | 0x20)) sign = 1;
	if (Delimiter(negative)) sign = -1;
	if (Delimiter(negative | 0x20)) sign = -1;

	if (!Double(deg)) return False();
	Delimiter('');
	Double(min);
	Delimiter('\'');
	Double(sec);
	Delimiter('"');

	if (Delimiter(positive)) sign = 1;
	if (Delimiter(positive | 0x20)) sign = 1;
	if (Delimiter(negative)) sign = -1;
	if (Delimiter(negative | 0x20)) sign = -1;

	if (deg<0) { sign = -sign; deg = -deg; }
	if ( (min<0) || (min>=60) || (sec<0) || (sec>=60) ) return False();
	
	deg = deg+min/60.0+sec/3600.0;
	if (deg>180) return False();

	degrees = sign*deg;
	return true;
}
