
#if !defined(FORMAT_H_INCLUDED)
#define FORMAT_H_INCLUDED

/// WIN32 DEFINITIONS
#if defined(WIN32)
	#define FORMAT
	#include <stdio.h>
	#include <string.h>
	#include <time.h>
	const int granularity = 16;
#endif

/// EPOC32 DEFINITIONS
#if defined(__SYMBIAN32__)
	#define FORMAT
	#include <e32std.h>
	#include <string.h>
	#include <time.h>
	const int granularity = 16;
#endif

/// PALMOS DEFINITIONS
#if defined(__palmos__)
	#include <PalmOS.h>
	#include "stdpalm.h"
	#define FORMAT SECTION(format)
	const int granularity = 16;
#endif

class CStr {
public:
	CStr() FORMAT;
	CStr(const char *str) FORMAT;
	CStr(const char ch) FORMAT;
	CStr(const CStr &str) FORMAT;
	~CStr() FORMAT;
	CStr& operator=(const char ch) FORMAT;
	CStr& operator=(const char *str) FORMAT;
	CStr& operator=(const CStr &str) FORMAT;
	CStr& operator+=(const char ch) FORMAT;
	CStr& operator+=(const char *str) FORMAT;
	CStr& operator+=(const CStr &str) FORMAT;
	CStr operator+(const char *str) const FORMAT;
	CStr operator+(const char ch) const FORMAT;
	CStr operator+(const CStr &str) const FORMAT;
	bool operator==(const CStr &str) const FORMAT;
	bool operator==(const char *str) const FORMAT;
	bool operator!=(const CStr &str) const FORMAT;
	bool operator!=(const char *str) const FORMAT;
	friend CStr operator+(const char *str1,const CStr &str2) FORMAT;
	const char* Ptr() const FORMAT;
	operator const char* () const FORMAT;
	char* DynamicCopy() const FORMAT;
	const int Length() const FORMAT;
	void PadL(int min, char ch) FORMAT;
	void Set(int start, int stop, char ch) FORMAT;
	void Remove(char ch) FORMAT;
private:
	int len,max;
	char *ptr;
	CStr(const int l) FORMAT;
	void Allocate(int atleast) FORMAT;
	void Reallocate(int atleast) FORMAT;
};

class CFormat {
public:
	static CStr SDouble(double number,int frac,int min=0,char pad=' ') FORMAT;
	static CStr SLong(long number,int min=0,char pad=' ') FORMAT;
	CStr Double(double number,int frac,int min=0,char pad=' ') FORMAT;
	CStr Long(long number,int min=0,char pad=' ') FORMAT;
};

typedef int DegFormat;
typedef int SpeedFormat;
typedef int DistFormat;
const DegFormat   degDDD = 0;
const DegFormat   degDMM = 1;
const DegFormat   degDMS = 2;
const SpeedFormat spdKPH = 0;
const SpeedFormat spdMPS = 1;
const DistFormat  dstMT  = 0;
const DistFormat  dstFT  = 1;

class CFormatGPS : public CFormat {
public:
	CFormatGPS() FORMAT;
	void Setup(DegFormat deg = degDMS, SpeedFormat speed = spdKPH,
	           DistFormat dist = dstMT, DistFormat alt = dstMT) FORMAT;
	CStr Latitude(double lat_d) FORMAT;
	CStr Longitude(double long_d) FORMAT;
	CStr Northing(double northing) FORMAT;
	CStr Easting(double easting) FORMAT;
	CStr Speed(double speed) FORMAT;
	CStr Dist(double dist) FORMAT;
	CStr EPE(double epe) FORMAT;
	CStr Time(time_t time) FORMAT;
	CStr Date(time_t time) FORMAT;
	CStr Altitude(double dist) FORMAT;
	CStr Course(double course) FORMAT;
protected:
	CStr Degrees(double degrees,char positive,char negative) FORMAT;
	DegFormat f_deg;
	SpeedFormat f_speed;
	DistFormat f_dist;
	DistFormat f_alt;
};

class CParse {
public:
	CParse(const CStr &str, char delimiter = ';') FORMAT;
	CParse(const char* str = "", char delimiter = ';') FORMAT;
	void Init(const CStr &str, char delimiter = ';') FORMAT;
	void Init(const char* str, char delimiter = ';') FORMAT;
	bool Delimiter() FORMAT;
	bool Delimiter(char delimiter) FORMAT;
	bool Character(char &ch) FORMAT;
	bool Const(const char *str) FORMAT;
	bool String(CStr &string) FORMAT;
	bool String(CStr &string, char delimiter) FORMAT;
	bool Long(int &number, int max = 0) FORMAT;
	bool Long(long &number, int max = 0) FORMAT;
	bool Hex(int &number, int max = 0) FORMAT;
	bool Hex(long &number, int max = 0) FORMAT;
	bool Double(float &number) FORMAT;
	bool Double(double &number) FORMAT;
	bool Skip(char delimiter) FORMAT;
	bool Skip() FORMAT;
	bool Ok() { return ok; };
	static bool Long(const char* str, long &number) FORMAT;
	static bool Long(const char* str, int &number) FORMAT;
	static bool Hex(const char* str, long &number) FORMAT;
	static bool Hex(const char* str, int &number) FORMAT;
protected:
	bool Advance(int count) FORMAT;
	void Spaces() FORMAT;
	bool Empty() { return (length==0); }
	bool False() { ok = false; return false; }
private:
	const char* ptr;
	int length;
	bool ok;
	char default_delimiter;
};

class CParseGPS : public CParse {
public:
	CParseGPS(const CStr &str, char delimiter = ';') : CParse(str, delimiter) {};
	CParseGPS(const char* str, char delimiter = ';') : CParse(str, delimiter) {};
	bool Latitude(double &lat_d) FORMAT;
	bool Longitude(double &long_d) FORMAT;
	static bool Latitude(const char* str, double &lat_d) FORMAT;
	static bool Longitude(const char* str, double &long_d) FORMAT;
protected:
	bool Degrees(double &degrees,char positive,char negative) FORMAT;
};

#endif
